@@ -173,6 +173,13 @@
     {}
 #endif
 
+#if BUILD_PRIMITIVE_ALL || BUILD_SDPA
+#define REG_SDPA_P(...) __VA_ARGS__
+#else
+#define REG_SDPA_P(...) \
+    { nullptr }
+#endif
+
 #if BUILD_PRIMITIVE_ALL || BUILD_SHUFFLE
 #define REG_SHUFFLE_P(...) __VA_ARGS__
 #else
@@ -238,4 +245,10 @@
 #define REG_XE2_ISA(...)
 #endif
 
+#if BUILD_PRIMITIVE_GPU_ISA_ALL || BUILD_XE3
+#define REG_XE3_ISA(...) __VA_ARGS__
+#else
+#define REG_XE3_ISA(...)
+#endif
+
 #endif
